from typing import Union

from cruisecontrolclient.client.CCParameter.Parameter import AbstractParameter


class AbstractBooleanParameter(AbstractParameter):
    def __init__(self, value: Union[bool, str]):
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) != bool and type(self.value) != str:
            raise ValueError(f"{self.value} is neither a boolean nor a string")


class AllowCapacityEstimationParameter(AbstractBooleanParameter):
    """allow_capacity_estimation=[true/false]"""
    name = 'allow_capacity_estimation'
    description = "Whether to allow capacity estimation when cruise-control is unable to obtain all per-broker capacity information"
    argparse_properties = {
        'args': ('--allow-capacity-estimation',),
        # Presume that the sensible default is to NOT allow capacity estimation
        'kwargs': dict(help=description, action='store_true')
    }


class ClearMetricsParameter(AbstractBooleanParameter):
    """clearmetrics=[true/false]"""
    name = 'clearmetrics'
    description = 'Clear the existing metric samples'
    argparse_properties = {
        'args': ('--clearmetrics', '--clear-metrics'),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class DryRunParameter(AbstractBooleanParameter):
    """dryRun=[true/false]"""
    # Docstrings list 'dryRun', but DRY_RUN_PARAM is 'dryrun'
    name = 'dryrun'
    description = 'Calculate, but do not execute, the cruise-control proposal'
    argparse_properties = {
        'args': ('-n', '--dry-run', '--dryrun'),
        # Presume that the lack of --dryrun affirmatively means no dry run
        'kwargs': dict(help=description, action='store_true')
    }


class ExcludeFollowerDemotionParameter(AbstractBooleanParameter):
    """exclude_follower_demotion=[true/false]"""
    name = 'exclude_follower_demotion'
    description = 'Whether to operate on partitions which only have follower replicas on the specified broker(s)'
    argparse_properties = {
        'args': ('--exclude-follower-demotion',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class ExcludeRecentlyDemotedBrokersParameter(AbstractBooleanParameter):
    """exclude_recently_demoted_brokers=[true/false]"""
    name = 'exclude_recently_demoted_brokers'
    description = 'Whether to exclude all recently-demoted brokers from this endpoint\'s action'
    argparse_properties = {
        'args': ('--exclude-recently-demoted-brokers', '--exclude-recently-demoted', '--exclude-demoted'),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class ExcludeRecentlyRemovedBrokersParameter(AbstractBooleanParameter):
    """exclude_recently_removed_brokers=[true/false]"""
    name = 'exclude_recently_removed_brokers'
    description = 'Whether to exclude all recently-removed brokers from this endpoint\'s action'
    argparse_properties = {
        'args': ('--exclude-recently-removed-brokers', '--exclude-recently-removed', '--exclude-removed'),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class IgnoreProposalCacheParameter(AbstractBooleanParameter):
    """ignore_proposal_cache=[true/false]"""
    name = 'ignore_proposal_cache'
    description = 'Whether to ignore the proposal cache'
    argparse_properties = {
        'args': ('--ignore-proposal-cache',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class JSONParameter(AbstractBooleanParameter):
    """json=[true/false]"""
    name = 'json'
    description = "Whether cruise-control's response should be in JSON format"
    argparse_properties = {
        'args': ('--json', '--json-response'),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class MaxLoadParameter(AbstractBooleanParameter):
    """max_load=[true/false]"""
    name = 'max_load'
    description = "Whether to return the peak load (max load) or the average load"
    argparse_properties = {
        'args': ('--max-load',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class SkipHardGoalCheckParameter(AbstractBooleanParameter):
    """skip_hard_goal_check=[true/false]"""
    name = 'skip_hard_goal_check'
    description = "Whether cruise-control should skip all hard goal checks"
    argparse_properties = {
        'args': ('--skip-hard-goal-check', '--skip-hard-goal-checks'),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class SkipRackAwarenessCheckParameter(AbstractBooleanParameter):
    """skip_rack_awareness_check=[true/false]"""
    name = 'skip_rack_awareness_check'
    description = "Whether to skip the rack-awareness check when performing this action"
    argparse_properties = {
        'args': ('--skip-rack-awareness-check',),
        # Presume that the sensible default is to NOT skip rack awareness
        'kwargs': dict(help=description, action='store_true')
    }


class SkipURPDemotionParameter(AbstractBooleanParameter):
    """skip_urp_demotion=[true/false]"""
    name = 'skip_urp_demotion'
    description = 'Whether to operate on partitions which are currently under-replicated'
    argparse_properties = {
        'args': ('--skip-urp-demotion',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class SuperVerboseParameter(AbstractBooleanParameter):
    """super_verbose=[true/false]"""
    name = 'super_verbose'
    description = "Whether cruise-control's response should return super-verbose information"
    argparse_properties = {
        'args': ('--super-verbose',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class ThrottleRemovedBrokerParameter(AbstractBooleanParameter):
    """throttle_removed_broker=[true/false]"""
    name = 'throttle_removed_broker'
    description = "Whether to apply the concurrent_partition_movements_per_broker limitation to the removed broker"
    argparse_properties = {
        'args': ('--throttle-removed-broker',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class UseReadyDefaultGoalsParameter(AbstractBooleanParameter):
    """use_ready_default_goals=[true/false]"""
    name = 'use_ready_default_goals'
    description = "Whether cruise-control should use its ready default goals"
    argparse_properties = {
        'args': ('--use-ready-default-goals',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }


class VerboseParameter(AbstractBooleanParameter):
    """verbose=[true/false]"""
    name = 'verbose'
    description = 'Whether cruise-control\'s response should return verbose information'
    argparse_properties = {
        'args': ('--verbose',),
        # There is no sensible default here; allow for ternary state
        'kwargs': dict(help=description, type=bool, metavar='BOOL')
    }



#!/usr/bin/env python3

# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

# To be able to easily parse command-line arguments
import argparse

# To be able to easily pass around the available endpoints and parameters
from cruisecontrolclient.client.ExecutionContext import ExecutionContext

# Convenience functions for displaying the retrieved Response
from cruisecontrolclient.client.Display import display_response

# To be able to instantiate Endpoint objects
import cruisecontrolclient.client.Endpoint as Endpoint

# To be able to make requests and get a response given a URL to cruise-control
from cruisecontrolclient.client.Responder import AbstractResponder, JSONDisplayingResponderGet, \
    JSONDisplayingResponderPost

# To be able to compose a URL to hand to requests
from cruisecontrolclient.client.Query import generate_url_from_cc_socket_address


def get_endpoint(args: argparse.Namespace,
                 execution_context: ExecutionContext) -> Endpoint.AbstractEndpoint:
    # Deepcopy the args so that we can delete keys from it as we handle them.
    #
    # Otherwise successive iterations will step on each other's toes.
    arg_dict = vars(args)

    # If we have a broker list, we need to make it into a comma-separated list
    # and pass it to the Endpoint at instantiation.
    if 'brokers' in arg_dict:
        comma_broker_id_list = ",".join(args.brokers)
        endpoint: Endpoint.AbstractEndpoint = execution_context.dest_to_Endpoint[args.endpoint_subparser](
            comma_broker_id_list)
        # Prevent trying to add this parameter a second time.
        del arg_dict['brokers']

    # Otherwise we can directly instantiate the Endpoint
    else:
        endpoint: Endpoint.AbstractEndpoint = execution_context.dest_to_Endpoint[args.endpoint_subparser]()

    # Iterate only over the parameter flags; warn user if conflicts exist
    for flag in arg_dict:
        if flag in execution_context.non_parameter_flags:
            pass
        else:
            # Presume None is ternary for ignore
            if arg_dict[flag] is not None:
                param_name = execution_context.flag_to_parameter_name[flag]
                # Check for conflicts in this endpoint's parameter-space,
                # which here probably means that the user is specifying more
                # than one irresolvable flag.
                #
                # For the StateEndpoint only, we don't care if we overwrite it.
                # This is because we presume 'substates:executor' at instantiation.
                if endpoint.has_param(param_name) and not isinstance(endpoint, Endpoint.StateEndpoint):
                    existing_value = endpoint.get_value(param_name)
                    raise ValueError(
                        f"Parameter {param_name}={existing_value} already exists in this endpoint.\n"
                        f"Unclear whether it's safe to remap to {param_name}={arg_dict[flag]}")
                else:
                    # If we have a destination broker list, we need to make it into a comma-separated list
                    if flag == 'destination_broker':
                        comma_broker_id_list = ",".join(arg_dict[flag])
                        endpoint.add_param(param_name, comma_broker_id_list)
                    else:
                        endpoint.add_param(param_name, arg_dict[flag])

    # We added this parameter already; don't attempt to add it again
    if 'destination_broker' in arg_dict:
        del arg_dict['destination_broker']

    # Handle hacking in json=true, if the user hasn't specified
    #
    # This is because it is easier to know when a JSON response is final,
    # compared to a text response.
    #
    # In fact, because it is not possible programmatically to know when a
    # text response is final, Responder actually does not support text responses.
    if not endpoint.has_param('json'):
        endpoint.add_param('json', 'true')

    # Handle add-parameter and remove-parameter flags
    #
    # Handle deconflicting adding and removing parameters, but don't
    # warn the user if they're overwriting an existing flag, since
    # these flags are meant as an admin-mode workaround to well-meaning defaults
    adding_parameter = 'add_parameter' in arg_dict and arg_dict['add_parameter']
    if adding_parameter:
        # Build a dictionary of parameters to add
        parameters_to_add = {}

        for item in arg_dict['add_parameter']:
            # Check that parameter contains an =
            if '=' not in item:
                raise ValueError("Expected \"=\" in the given parameter")

            # Check that the parameter=value string is correctly formatted
            split_item = item.split("=")
            if len(split_item) != 2:
                raise ValueError("Expected only one \"=\" in the given parameter")
            if not split_item[0]:
                raise ValueError("Expected parameter preceding \"=\"")
            if not split_item[1]:
                raise ValueError("Expected value after \"=\" in the given parameter")

            # If we are here, split_item is a correctly-formatted list of 2 items
            parameter, value = split_item
            # Add it to our running dictionary
            parameters_to_add[parameter] = value

    # The 'remove_parameter' string may not be in our namespace, and even if it
    # is, there may be no parameters supplied to it.
    #
    # Accordingly, check both conditions and store as a simpler boolean.
    removing_parameter = 'remove_parameter' in arg_dict and arg_dict['remove_parameter']
    if removing_parameter:
        # Build a set of parameters to remove
        parameters_to_remove = set()
        for item in arg_dict['remove_parameter']:
            parameters_to_remove.add(item)

    # Validate that we didn't receive ambiguous input
    if adding_parameter and removing_parameter:
        if set(parameters_to_add) & parameters_to_remove:
            raise ValueError("Parameter present in --add-parameter and in --remove-parameter; unclear how to proceed")

    # Having validated parameters, now actually add or remove them.
    #
    # Do this without checking for conflicts from existing parameter=value mappings,
    # since we presume that if the user supplied these, they really want them
    # to override existing parameter=value mappings
    if adding_parameter:
        for parameter, value in parameters_to_add.items():
            endpoint.add_param(parameter, value)
    if removing_parameter:
        for parameter in parameters_to_remove:
            endpoint.remove_param(parameter)

    return endpoint


def get_responder(endpoint: Endpoint.AbstractEndpoint,
                  url: str) -> AbstractResponder:
    # Handle instantiating the correct Responder
    if endpoint.http_method == "GET":
        json_responder = JSONDisplayingResponderGet(url)
    elif endpoint.http_method == "POST":
        json_responder = JSONDisplayingResponderPost(url)
    else:
        raise ValueError(f"Unexpected http_method {endpoint.http_method} in endpoint")

    return json_responder


def build_argument_parser(execution_context: ExecutionContext) -> argparse.ArgumentParser:
    """
    Builds and returns an argument parser for interacting with cruise-control via CLI.

    It is expected that you can substitute another function for this function
    that returns a parser which is decorated similarly.

    :return:
    """

    # Define some inner functions that make no sense outside of this context
    def add_add_parameter_argument(p: argparse.ArgumentParser):
        """
        This should be used with all cruise-control endpoint parsers, to provide
        forward compatibility and greater operational flexibility.
        :param p:
        :return:
        """
        p.add_argument('--add-parameter', '--add-parameters', metavar='PARAM=VALUE',
                       help="Manually specify one or more parameter and its value in the cruise-control endpoint, "
                            "like 'param=value'",
                       nargs='+')
        execution_context.non_parameter_flags.add('add_parameter')

    def add_remove_parameter_argument(p: argparse.ArgumentParser):
        """
        Adds the ability to manually specify parameters to remove from the cruise-control
        endpoint, of the form 'parameter'.

        This should be used with all cruise-control endpoint parsers, to provide
        forward compatibility and greater operational flexibility.

        :param p:
        :return:
        """
        p.add_argument('--remove-parameter', '--remove-parameters', metavar='PARAM',
                       help="Manually remove one or more parameter from the cruise-control endpoint, like 'param'",
                       nargs='+')
        execution_context.non_parameter_flags.add('remove_parameter')

    # Display command-line arguments for interacting with cruise-control
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--socket-address', help="The hostname[:port] of the cruise-control to interact with",
                        required=True)
    execution_context.non_parameter_flags.add('socket_address')

    # Define subparser for the different cruise-control endpoints
    #
    # "dest" allows the name of the subparser used to be stored at args.endpoint_subparser,
    # which for some weird reason doesn't happen otherwise.
    endpoint_subparser = parser.add_subparsers(title='endpoint',
                                               description='Which cruise-control endpoint to interact with',
                                               # 'endpoint' would collide with an existing cc parameter
                                               dest='endpoint_subparser')
    execution_context.non_parameter_flags.add('endpoint_subparser')

    # A map from endpoint names to that endpoint's argparse parser
    endpoint_to_parser_instance = {}

    # Dynamically build an argparse CLI from the Endpoint and Parameter properties
    for endpoint in execution_context.available_endpoints:
        endpoint_parser = endpoint_subparser.add_parser(*endpoint.argparse_properties['args'],
                                                        **endpoint.argparse_properties['kwargs'])
        endpoint_to_parser_instance[endpoint.name] = endpoint_parser
        for parameter in endpoint.available_Parameters:
            endpoint_parser.add_argument(*parameter.argparse_properties['args'],
                                         **parameter.argparse_properties['kwargs'])
        # Hack in some future-proofing by allowing users to add and remove parameter=value mappings
        add_add_parameter_argument(endpoint_parser)
        add_remove_parameter_argument(endpoint_parser)

    return parser


def main():
    # Instantiate a convenience class to pass around information about available endpoints and parameters.
    e = ExecutionContext()

    # Display and parse command-line arguments for interacting with cruise-control
    parser = build_argument_parser(e)
    args = parser.parse_args()

    # Get the endpoint that the parsed args specify
    endpoint = get_endpoint(args=args, execution_context=e)

    # Get the socket address for the cruise-control we're communicating with
    cc_socket_address = args.socket_address

    # Generate the correct URL from the endpoint and the socket address
    url = generate_url_from_cc_socket_address(cc_socket_address=cc_socket_address,
                                              endpoint=endpoint)

    # Get a responder from the given URL and endpoint
    json_responder = get_responder(endpoint=endpoint, url=url)

    # Retrieve the response and display it
    response = json_responder.retrieve_response()
    display_response(response)


if __name__ == "__main__":
    main()



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

# To be able to make more-precise type hints
from cruisecontrolclient.client.Endpoint import AbstractEndpoint


def generate_url_from_cc_socket_address(cc_socket_address: str, endpoint: AbstractEndpoint) -> str:
    """
    Given a cruise-control hostname[:port] and an Endpoint, return the correct URL
    for this cruise-control operation.

    :param cc_socket_address: like hostname[:port], ip-address[:port]
    :param endpoint:
    :return: URL, the correct URL to perform the Endpoint's operation
             on the given cruise-control host.
    """
    url = f"http://{cc_socket_address}/kafkacruisecontrol/{endpoint.compose_endpoint()}"
    return url



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

# Use convenience function to redirect printing to stderr
from cruisecontrolclient.util.print import print_error

# To allow for more precise type hints
from requests import Response

# To print the JSONs
from pprint import pformat

# To be able to show the humans the JSON in tabular form
from pandas import DataFrame, set_option

# To be able to give more-precise type hints
from typing import Dict, Callable  # noqa

# To be able to load a JSON from a string
import json

# To be able to indent the correct level for conveying JSON descent
import textwrap


def get_key_to_display_function() -> Dict[str, Callable]:
    """
    Convenience mapping of the key we see to the function that should parse its contents.

    :return: a mapping of keys to expected functions
    """
    return {
        'AnomalyDetectorState': display_anomaly_detector_state,
        'brokers': display_brokers,
        'goalReadiness': display_goal_readiness,
        'goalSummary': display_goal_summary,
        'hosts': display_hosts,
        'loadAfterOptimization': display_load_after_optimization,
        'KafkaBrokerState': display_kafka_broker_state,
        'KafkaPartitionState': display_kafka_partition_state,
        'progress': display_progress,
        'records': display_records
    }


def print_with_indent(s: str, level=0) -> None:
    """
    Prints the string with a level of indentation

    :param s:
    :param level:
    :return:
    """
    print_error(textwrap.indent(s, "\t" * level))


def display_anomaly_detector_state(AnomalyDetectorState_json: dict, level=0) -> None:
    """
    Displays the 'AnomalyDetectorState' key in a cruise-control response

    :param AnomalyDetectorState_json: the JSON structure referenced by the
            'AnomalyDetectorState' key.
    :return:
    """
    for key in AnomalyDetectorState_json:
        if key == "recentGoalViolations":
            print_with_indent(f"\t'{key}:", level)
            for item in AnomalyDetectorState_json[key]:
                if 'optimizationResult' in item:
                    optimization_json = json.loads(item['optimizationResult'])
                    del item['optimizationResult']
                    print_with_indent(pformat(item), level + 1)
                    print_with_indent("'optimizationResult':", level + 2)
                    display_dict(optimization_json, level + 2)

                else:
                    print_with_indent(pformat(item), level + 1)
        else:
            print_error(f"\t'{key}: {AnomalyDetectorState_json[key]}")


def display_brokers(brokers_list: list, level=0) -> None:
    """
    Displays the 'brokers' key in a cruise-control response

    :param brokers_list: the list structure referenced by the 'brokers' key.
    :return:
    """
    df = DataFrame(brokers_list)
    df.set_index('Broker', inplace=True)
    print_with_indent(df.to_string(), level)


def display_hosts(hosts_list: list, level=0) -> None:
    """
    Displays the 'hosts' key in a cruise-control response

    :param hosts_list: the list structure referenced by the 'hosts' key.
    :return:
    """
    df = DataFrame(hosts_list)
    df.set_index('Host', inplace=True)
    print_with_indent(df.to_string(), level)


def display_load_after_optimization(loadAfterOptimization_json: dict, level=0) -> None:
    """
    Displays the 'loadAfterOptimization' key in a cruise-control response

    :param loadAfterOptimization_json: the JSON structure referenced by the
            'loadAfterOptimization' key.
    :return:
    """
    # Make and print the dataframe from the 'brokers' subset of the JSON
    df = DataFrame(loadAfterOptimization_json['brokers'])
    df.set_index('Broker', inplace=True)
    print_with_indent(df.to_string(), level)


def display_kafka_broker_state(KafkaBrokerState_json: dict, level=0) -> None:
    """
    Displays the 'KafkaBrokerState' key in a cruise-control response

    :param KafkaBrokerState_json: the JSON structure referenced by the
            'KafkaBrokerState' key.
    :return:
    """
    # Handle list-like values, which we expect for 'OnlineLogDirsByBrokerId' and
    # 'OfflineLogDirsByBrokerId'
    log_dir_df = None
    # Items under these keys should be like
    # '33967': []
    # '33967': ['/path/to/only/logdir']
    # '33967': ['/path/to/one/logdir', /path/to/another/logdir', ...]
    if 'OnlineLogDirsByBrokerId' in KafkaBrokerState_json and 'OfflineLogDirsByBrokerId' in KafkaBrokerState_json:

        # Create a dictionary of the comma-joined log-dirs
        broker_id_to_log_dir = {}
        for entry in ('OfflineLogDirsByBrokerId', 'OnlineLogDirsByBrokerId'):
            broker_id_to_log_dir[entry] = {key: ','.join(value) for key, value in
                                           KafkaBrokerState_json[entry].items()}
            # Delete from the original dict, since we later presume int32 type,
            # and need to preserve that assumption
            del KafkaBrokerState_json[entry]

        # Create a DataFrame of this logDir info.
        # We will join this to the main DataFrame once it is constructed.
        log_dir_df: DataFrame = DataFrame(broker_id_to_log_dir)
        log_dir_df.index.name = "BrokerId"

    # KafkaBrokerState_json should contain 'LeaderCountByBrokerId', 'OutOfSyncCountByBrokerId', and
    # 'ReplicaCountByBrokerId', each of which should map '{brokerID}': count
    df: DataFrame = DataFrame(KafkaBrokerState_json, dtype=int)
    df.fillna(0, inplace=True)
    df = df.astype('int32')
    df.index.name = "BrokerId"
    if log_dir_df is not None:
        df = df.join(log_dir_df)
    print_with_indent(df.to_string(), level)


def display_kafka_partition_state(KafkaPartitionState_json: dict, level=0) -> None:
    """
    Displays the 'KafkaPartitionState' key in a cruise-control response

    :param KafkaPartitionState_json: the JSON structure referenced by the
            'KafkaPartitionState' key.
    :return:
    """
    # We expect KafkaPartitionState_json to contain 'offline': [...] and 'urp': [...]
    for key in KafkaPartitionState_json:
        if type(KafkaPartitionState_json[key] == list):
            print_with_indent(f"\t'{key}'", level)
            if KafkaPartitionState_json[key]:
                # Reuse the topic-partition dataframe display function
                display_records(KafkaPartitionState_json[key], level)
            else:
                # If we're here, there's an empty list
                print_with_indent(str(KafkaPartitionState_json[key]), level)


def display_goal_readiness(goalReadiness_list: list, level=0) -> None:
    """
    Displays the 'goalReadiness' key in a cruise-control response.

    Note that this is not a top-level key, but is rather a sub-key of the
    AnalyzerState top-level key.

    :param goalReadiness_list: the list structure referenced by the
            'goalReadiness' key.
    :return:
    """
    for elem in goalReadiness_list:
        print_with_indent(f"{elem['name']}: {elem['status']}", level)
        print_with_indent(pformat(elem['modelCompleteRequirement']))
        print_error()


def display_goal_summary(goalSummary_json: dict, level=0) -> None:
    """
    Displays the 'goalSummary' key in a cruise-control response

    :param goalSummary_json: the JSON structure referenced by the 'goalSummary' key.
    :return:
    """
    # Make and print a dataframe for each of the goals in clusterModelState
    for elem in goalSummary_json:
        print_with_indent(f"{elem['goal']}: {elem['status']}", level)
        df = DataFrame(elem['clusterModelStats']['statistics'])
        print_with_indent(df.to_string(), level)
        print_error()


def display_progress(progress_list: list, level=0) -> None:
    """
    Displays the 'progress' key in a (non-final) cruise-control response

    :param progress_list: the list structure referenced by the 'progress' key.
    :return:
    """
    for elem in progress_list:
        print_error(f"operation: {elem['operation']}")
        df = DataFrame(elem['operationProgress'])
        try:
            df.set_index('step', inplace=True)
        # The dataframe we get back may not have 'step' as a key
        except KeyError:
            pass
        df_string = df.to_string()
        # Indent the dataframe for better display
        df_string = df_string.replace("\n", "\n    ")
        print_with_indent(df_string, level)

    # Print ellipses so the humans know we are waiting :)
    print_error(".........")


def display_records(records_list: list, level=0) -> None:
    """
    Displays the 'records' key in a cruise-control response

    :param records_list: the list structure referenced by the 'records' key.
    :return:
    """
    df = DataFrame(records_list)
    df.set_index(['topic', 'partition'], inplace=True)
    df.sort_index(inplace=True)
    print_with_indent(df.to_string(), level)


def display_response(response: Response) -> None:
    """
    Handles setting the display options and extracting the dict from the response,
    then beginning the (possibly recursive) display parsing.

    :param response:
    :return:
    """
    # Set pandas display options
    # Display floats with a ',' 1000s separator, to two decimal places
    set_option('display.float_format', "{:,.2f}".format)
    # Display all rows and columns in the dataframe.  Don't leave any out
    set_option('display.max_rows', int(1E12))
    set_option('display.max_columns', int(1E12))
    # Display the full column width, even if it's very wide.  Don't truncate it.
    set_option('display.max_colwidth', 10000)
    # Don't wrap the table we show
    set_option('display.expand_frame_repr', False)
    # Don't 'sparsify' sorted multi-index dataframes, since
    # sparsifying makes it harder to do bash-type processing of tabular data
    set_option('display.multi_sparse', False)

    j: dict = response.json()

    try:
        display_dict(j)
    except Exception as e:
        print_error(f"Unhandled exception during display; showing raw JSON", e)
        print_error(pformat(j))


def display_dict(j: dict, level=0) -> None:
    """
    Walks the cruise-control Response dict and attempts to display
    the given top-level keys to the humans.


    :param j:
    :return:
    """
    for key in j.keys():
        print_with_indent(f"'{key}':", level)
        # Display the contents of this key with its key-specific function, if possible.
        key_to_display_function = get_key_to_display_function()
        if key in key_to_display_function:
            key_to_display_function[key](j[key], level)
        # Otherwise, just pretty-print this key's contents.
        else:
            print_with_indent(pformat(j[key]), level)
        print_error()



import re
from typing import Union

from cruisecontrolclient.client.CCParameter.Parameter import AbstractParameter


class AbstractRegularExpressionParameter(AbstractParameter):
    def __init__(self, value: Union[str, re.Pattern]):
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) == re.Pattern:
            # If we've been passed a re.Pattern, obtain the string that comprised
            # this re.Pattern, and store that string as our value.
            #
            # This allows us to store a value that is the closest to what
            # cruise-control expects.
            self.value = self.value.pattern
        elif type(self.value) == str:
            # A string is not necessarily a regular expression; verify whether
            # it compiles into one.
            try:
                re.compile(self.value)
            except re.error as e:
                raise ValueError(f"{self.value} is not a valid regular expression", e)
        else:
            raise ValueError(f"{self.value} must be either a string or a re.Pattern")


class ExcludedTopicsParameter(AbstractRegularExpressionParameter):
    """excluded_topics=[pattern]"""
    name = 'excluded_topics'
    description = 'A regular expression matching which topics to exclude from this endpoint\'s action'
    argparse_properties = {
        'args': ('--excluded-topics', '--exclude-topics', '--exclude-topic'),
        'kwargs': dict(help=description, metavar='REGEX')
    }


class TopicParameter(AbstractRegularExpressionParameter):
    """topic=[topic]"""
    name = 'topic'
    description = 'A regular expression matching the desired topics'
    argparse_properties = {
        'args': ('--topic',),
        'kwargs': dict(help=description, metavar='REGEX')
    }



from cruisecontrolclient.client.CCParameter.Parameter import AbstractParameter


class AbstractNonNegativeIntegerParameter(AbstractParameter):
    def __init__(self, value: int):
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) != int:
            raise ValueError(f"{self.value} is not an integer value")
        elif self.value < 0:
            raise ValueError(f"{self.value} must be a non-negative integer")


class ReviewIDParameter(AbstractNonNegativeIntegerParameter):
    """review_id=[id]"""
    name = 'review_id'
    description = 'The id of the approved review'
    argparse_properties = {
        'args': ('--review-id',),
        'kwargs': dict(metavar='K', help=description, type=int)
    }


from typing import Union

from cruisecontrolclient.client.CCParameter.Parameter import AbstractParameter


class AbstractTimestampParameter(AbstractParameter):
    """
    For parameters that accept timestamps in milliseconds since the Epoch.
    """

    def __init__(self, value: Union[str, int]):
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        """
        Although there are many possible validations against a timestamp, none
        are satisfactorily one-size-fits-all.

        Accordingly, only type checking will be performed against the given value.
        :return:
        """
        if type(self.value) == str:
            try:
                int(self.value)
            except ValueError as e:
                raise ValueError(f"{self.value} cannot be cast to integer timestamp", e)
        elif type(self.value) != int:
            raise ValueError(f"{self.value} must be either a string or an integer")


class EndParameter(AbstractTimestampParameter):
    """start=[END_TIMESTAMP]"""
    name = 'end'
    description = 'The timestamp in milliseconds since the Epoch to use as the end'
    argparse_properties = {
        'args': ('--end-timestamp',),
        'kwargs': dict(help=description, metavar='END_TIMESTAMP')
    }


class StartParameter(AbstractTimestampParameter):
    """start=[START_TIMESTAMP]"""
    name = 'start'
    description = 'The timestamp in milliseconds since the Epoch to use as the start'
    argparse_properties = {
        'args': ('--start-timestamp',),
        'kwargs': dict(help=description, metavar='START_TIMESTAMP')
    }


class TimeParameter(AbstractTimestampParameter):
    """time=[TIMESTAMP]"""
    name = 'time'
    description = 'The timestamp in milliseconds since the Epoch to use'
    argparse_properties = {
        'args': ('--timestamp',),
        'kwargs': dict(help=description, metavar='TIMESTAMP')
    }



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

import setuptools

LONG_DESCRIPTION ='''`cruise-control-client` is an API-complete Python client for [`cruise-control`](https://github.com/linkedin/cruise-control).

It comes with a command-line interface to the client (`cccli`) (see [README](https://github.com/linkedin/cruise-control/tree/master/docs/wiki/Python%20Client)).

`cruise-control-client` can also be used in Python applications needing programmatic access to `cruise-control`.'''

setuptools.setup(
    name='cruise-control-client',
    version='0.1.4',
    author='mgrubent',
    author_email='mgrubentrejo@linkedin.com',
    description='A Python client for cruise-control',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/linkedin/cruise-control',
    entry_points={
        'console_scripts': [
            'cccli = cruisecontrolclient.client.cccli:main'
        ]
    },
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas',
        'requests'
    ],
    license='Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").'
            ' See License in the project root for license information.',
)
